<?php

/**
 * @file
 * Drush commands for Profile Lite module.
 */

/**
 * Implementation of hook_drush_help().
 */
function profile_lite_drush_help($section) {
  switch ($section) {
    case 'drush:content-profile-migrate':
      return dt('Migrate Drupal 6 Content Profile nodes to user fields. Using ' .
                'this command on a large set of profiles is not recommended. '  .
                'Instead, use Migrate module.');
  }
}

/**
 * Implementation of hook_drush_command().
 */
function profile_lite_drush_command() {
  $items['content-profile-migrate'] = array(
    'description' => 'Migrate Drupal 6 Content Profile nodes to user fields.',
    'drupal dependencies' => array('field_tools', 'content_migrate'),
    'options' => array(
      'timers' => 'Display timers upon completion.',
    ),
  );

  return $items;
}

/**
 * Command callback for content-profile-migrate.
 *
 * TODO: This command really needs to use Queue API, but for smaller sets of
 * data it works fine.
 */
function drush_profile_lite_content_profile_migrate() {
  module_load_include('admin.inc', 'field_tools');
  module_load_include('inc', 'content_migrate', 'includes/content_migrate.admin');

  // Load up content migrate options, so we can make sure that all data has been
  // migrated from CCK to Drupal 7 fields.
  $content_migrate_fields = content_migrate_get_options();
  if (empty($content_migrate_fields['converted'])) {
    drush_log(dt("It appears that CCK's Content Migrate module has not been run " .
                 'yet. Please run drush content-migrate-fields prior to running ' .
                 'this script.'), 'error');
    return;
  }

  // Load all the node type names.
  $types = _node_types_build()->names;

  // Array of totals for all profile types, used in dt().
  $totals = array(
    '!updated' => 0,
    '!total' => 0,
    '!deleted' => 0,
    '!failed' => 0,
  );

  // Iterate over each type, checking to see if any of them are content_profiles.
  foreach ($types as $type => $name) {
    // Don't process node types that weren't from content_profile.
    if (!variable_get('content_profile_use_' . $type, FALSE)) {
      unset($types[$type]);
      continue;
    }

    // Set up some dt() placeholders.
    $dt_args = array(
      '!updated' => 0,
      '!total' => 0,
      '!deleted' => 0,
      '!failed'  => 0,
      '@type' => "$name ($type)",
    );

    drush_log(dt('Analyzing fields on node type @type.', $dt_args), 'status');
    // Get a fresh copy of the user fields for each type.
    $user_fields = field_info_instances('user', 'user');
    $instances = field_info_instances('node', $type);

    // Check for fields that have not been migrated by Content Migrate module.
    if ($unmigrated = array_diff_key($instances, $content_migrate_fields['converted'])) {
      $dt_args['@unmigrated'] = implode(', ', array_keys($unmigrated));
      drush_log(dt('The @unmigrated field(s) on node type @type have not been ' .
                   'migrated by Content Migrate module. This must be resolved ' .
                   'before the node type can be migrated to the User entity.', $dt_args), 'error');
      continue;
    }

    // Check to see if we need to create the profile_lite category.
    $question = 'There does not appear to be a profile category for @type. Would you like to create it?';
    if (module_exists('profile_lite') && !_profile_lite_category_exists($type) && drush_confirm(dt($question, $dt_args))) {
      $category = new stdClass();
      $category->name = $type;
      $category->label = $name;
      foreach ($instances as $field_name => $instance) {
        $category->fields[$field_name] = array('weight' => $instance['weight']);
      }
      profile_lite_category_save($category);
    }

    // Add the node type field instances to the user entity.
    if ($missing_instances = array_diff_key($instances, $user_fields)) {
      drush_log(dt('Saving field instances from @type to the user entity.', $dt_args), 'status');
      foreach ($missing_instances as $instance) {
        _field_tools_add_instance_to_bundles($instance, array('user' => array('user')));
      }
    }
    else {
      drush_log(dt('The user entity already contains all field instances from type @type. Continuing.', $dt_args), 'status');
    }

    // Load up the nodes of this type.
    $node_query = new EntityFieldQuery();
    $node_query->entityCondition('entity_type', 'node')
               ->entityCondition('bundle', $type);
    $result = $node_query->execute();
    // Well drat, we need the uid so we can't use this. Have to load the full nodes.
    //field_attach_load('node', $result['node']);
    drush_log(dt('Loading all nodes of type @type. This may take a while.', $dt_args), 'status');

    // If we don't have any nodes, no reason to continue. Move to the next type.
    if (empty($result['node'])) {
      drush_log(dt('There are not any nodes of this type. Continuing...'), 'status');
      continue;
    }

    // Now load up the nodes. If there are a lot of nodes, this will probably die,
    // but it works for smaller sets for now.
    $nodes = node_load_multiple(array_keys($result['node']));
    $totals['!total'] += $dt_args['!total'] = count($nodes);

    $question = 'Would you like to migrate !total total nodes of type @type to user entities?';
    if (!drush_confirm(dt($question, $dt_args))) {
      continue;
    }

    // Ask some questions up front so the user can set it and forget it.
    $msg = 'You have the option of overwriting any existing profile data that '      .
           'might exist in the Drupal 7 user fields with data from these content '   .
           'profile nodes. If this is a recent upgrade to Drupal 7, it is unlikely ' .
           'there is any data in these fields. If this is the case or you are sure ' .
           'the fields from the content_profile node types have never existed on '   .
           'the user entity before, it is safe to respond yes to the following '     .
           'question. If you are NOT sure, respond no.';
    drush_log(dt($msg), 'warning');
    $question = 'Would you like to overwrite existing field data on the user ' .
                'entities with data from node type @type?';
    $overwrite = drush_confirm(dt($question, $dt_args));

    $i = 0;
    $total = count($nodes);

    foreach ($nodes as $node) {
      timer_start(dt('migrate node of type @type', $dt_args));
      drush_print_progress(++$i / $total);

      $account = user_load($node->uid);
      $dt_args['@name'] = $account->name;
      $dt_args['@uid']  = $account->uid;
      // Remove everything but the profile fields from the node object.
      $profile_data = array_intersect_key((array) $node, $instances);
      // If we are preserving existing data, check for it.
      if (!$overwrite) {
        foreach (array_keys($instances) as $field_name) {
          if (!empty($account->$field_name)) {
            unset($profile_data[$field_name]);
          }
        }
      }
      if (empty($profile_data)) {
        drush_log(dt('No profile data to save for user @name (uid @uid).', $dt_args), 'status');
        timer_stop(dt('migrate node of type @type', $dt_args));
        continue;
      }
      if (user_save($account, $profile_data)) {
        drush_log(dt('Saved profile data for user @name (uid @uid).', $dt_args));
        $dt_args['!updated']++;
        $nodes_to_delete[] = $node->nid;
      }
      else {
        drush_log(dt('Failed to save profile data for user @name (uid @uid).', $dt_args), 'error');
        $dt_args['!failed']++;
      }
      timer_stop(dt('migrate node of type @type', $dt_args));
    }
    drush_log(dt('Migrated !updated of !total @type nodes.', $dt_args), 'success');
    if (!empty($dt_args['!failed'])) {
      drush_log(dt('Failed to migrate !failed nodes.', $dt_args), 'warning');
    }

    $question = 'Would you like to delete !delete-total nodes of type @type?';
    $dt_args['!delete-total'] = $total = count($nodes_to_delete);

    if ($nodes_to_delete && drush_confirm(dt($question, $dt_args))) {
      $i = 0;
      foreach ($nodes_to_delete as $nid) {
        timer_start(dt('delete node of type @type', $dt_args));
        node_delete($nid);
        $i++;
        $dt_args['!deleted'] = $i;
        drush_print_progress($i / $total);
        timer_stop(dt('delete node of type @type', $dt_args));
      }
      drush_log(dt('Deleted !deleted node(s).', $dt_args), 'success');
    }
    $totals['!updated'] += $dt_args['!updated'];
    $totals['!failed']  += $dt_args['!failed'];
    $totals['!deleted'] += $dt_args['!deleted'];
  }

  // Finished! Tally the score.
  $totals['!types-count'] = count($types);
  $totals['@types'] = implode(',', $types);
  $msg = 'Migrated !updated of !total node(s) from !types-count content_profile ' .
         'node type(s). !failed failed. Deleted !deleted node(s).';
  drush_log(dt($msg, $totals), 'status');
  if (drush_get_option('timers')) {
    drush_print_timers();
  }
}

// Trying to get these functions in core: http://drupal.org/node/1744410
if (!function_exists('drush_print_progress')) {
  /**
   * Print out progress based on ratio of completion.
   *
   * @param float $ratio
   *   The decimal ratio of completion. Do not round this value.
   * @param bool $print_when_verbose
   *   Whether to print even when in verbose or debug mode. Defaults to FALSE.
   */
  function drush_print_progress($ratio, $print_when_verbose = FALSE) {
    // Only print progress if not in quiet, verbose or debug mode.
    $verbose = drush_get_context('DRUSH_VERBOSE') || drush_get_context('DRUSH_DEBUG');

    // If the function is being called in quiet mode, or we're in verbose/debug
    // mode, simply return without printing anything.
    if (drush_get_context('DRUSH_QUIET') || ($verbose && !$print_when_verbose)) {
      return;
    }
    // \r is not supported by drush_backend, so we call a different function.
    drush_get_context('DRUSH_BACKEND') ? _drush_print_backend_progress($ratio) : _drush_print_progress($ratio);
  }

  /**
   * Prints out a progress bar directly to STDOUT. Do not call this function
   * directly, rather call drush_print_progress().
   */
  function _drush_print_progress($ratio) {
    $percentage = floor($ratio * 100) . '%';
    $columns = drush_get_context('DRUSH_COLUMNS', 80);
    // Subtract 8 characters for the percentage, brackets, spaces and arrow.
    $progress_columns = $columns - 8;
    // If ratio is 1 (complete), the > becomes a = to make a full bar.
    $arrow = ($ratio < 1) ? '>' : '=';
    // Print a new line if ratio is 1 (complete). Otherwise, use a CR.
    $line_ending = ($ratio < 1) ? "\r" : "\n";

    // Determine the current length of the progress string.
    $current_length = floor($ratio * $progress_columns);
    $progress_string = str_pad('', $current_length, '=');

    $output  = '[';
    $output .= $progress_string . $arrow;
    $output .= str_pad('', $progress_columns - $current_length);
    $output .= ']';
    $output .= str_pad('', 5 - strlen($percentage)) . $percentage;
    $output .= $line_ending;

    print $output;
  }

  /**
   * Logs a percentage progress every 10 seconds, for use when a command is called
   * in backend mode. Do not call this function directly, rather call
   * drush_print_progress().
   */
  function _drush_print_backend_progress($ratio) {
    static $next_print;
    $percentage = round($ratio * 100) . '%';
    $interval = 10;
    $time = time();

    if (!isset($next_print)) {
      $next_print = $time + $interval;
    }
    if ($ratio == 1 || $time >= $next_print) {
      $next_print += $interval;
      $dt_args['@percent'] = $percentage;
      drush_log(dt('@percent completed.', $dt_args), 'status');
    }
  }
}